from __future__ import annotations

import logging

from dynamicprompts import constants
from dynamicprompts.generators.promptgenerator import PromptGenerator

logger = logging.getLogger(__name__)


class BatchedCombinatorialPromptGenerator(PromptGenerator):
    def __init__(self, generator: PromptGenerator, batches: int = 1) -> None:
        self._generator = generator
        self._batches = batches

    def generate(
        self,
        template: str,
        max_prompts: int = constants.MAX_IMAGES,
    ) -> list[str]:
        images = []

        for _ in range(self._batches):
            images.extend(self._generator.generate(template, max_prompts))
        return images



from __future__ import annotations

import warnings

from dynamicprompts.wildcardmanager import WildcardManager


class ActionBuilder:  # pragma: no cover
    def __init__(self, wildcard_manager: WildcardManager, ignore_whitespace=False):
        warnings.warn(
            "ActionBuilder is deprecated and will be removed in a future version",
            DeprecationWarning,
        )
        self._wildcard_manager = wildcard_manager
        self._ignore_whitespace = ignore_whitespace



from __future__ import annotations

import logging

from dynamicprompts.commands import (
    Command,
    LiteralCommand,
    SequenceCommand,
    VariantCommand,
    WildcardCommand,
)
from dynamicprompts.sampling_context import SamplingContext
from dynamicprompts.types import StringGen
from dynamicprompts.utils import rotate_and_join

logger = logging.getLogger(__name__)


class Sampler:
    def generator_from_command(
        self,
        command: Command,
        context: SamplingContext,
    ) -> StringGen:
        # This is purposely not a dict lookup/getattr magic thing, to make
        # it easier for code completion etc. to see what's going on.
        if isinstance(command, LiteralCommand):
            return self._get_literal(command, context)
        if isinstance(command, SequenceCommand):
            return self._get_sequence(command, context)
        if isinstance(command, VariantCommand):
            return self._get_variant(command, context)
        if isinstance(command, WildcardCommand):
            return self._get_wildcard(command, context)
        return self._unsupported_command(command)

    def _unsupported_command(self, command: Command) -> StringGen:
        raise NotImplementedError(
            f"{self.__class__.__name__} does not support {command.__class__.__name__}",
        )

    def _get_wildcard(
        self,
        command: WildcardCommand,
        context: SamplingContext,
    ) -> StringGen:
        return self._unsupported_command(command)

    def _get_variant(
        self,
        command: VariantCommand,
        context: SamplingContext,
    ) -> StringGen:
        return self._unsupported_command(command)

    def _get_sequence(
        self,
        command: SequenceCommand,
        context: SamplingContext,
    ) -> StringGen:
        sub_generators = [context.generator_from_command(c) for c in command.tokens]

        while True:
            yield rotate_and_join(sub_generators, separator=command.separator)

    def _get_literal(
        self,
        command: LiteralCommand,
        context: SamplingContext,
    ) -> StringGen:
        while True:
            yield command.literal



from __future__ import annotations

from typing import Generator, Iterable, List

from dynamicprompts.commands import Command

StringGen = Generator[str, None, None]
CommandList = List[Command]
CommandListGen = Generator[CommandList, None, None]
StringIter = Iterable[str]


def to_string_gen(strs: StringIter) -> StringGen:
    yield from strs



from __future__ import annotations

import logging

from dynamicprompts import constants
from dynamicprompts.enums import SamplingMethod
from dynamicprompts.generators.promptgenerator import PromptGenerator
from dynamicprompts.parser.config import ParserConfig, default_parser_config
from dynamicprompts.sampling_context import SamplingContext
from dynamicprompts.wildcardmanager import WildcardManager

logger = logging.getLogger(__name__)


class CombinatorialPromptGenerator(PromptGenerator):
    def __init__(
        self,
        wildcard_manager: WildcardManager | None = None,
        ignore_whitespace: bool = False,
        parser_config: ParserConfig = default_parser_config,
    ) -> None:
        wildcard_manager = wildcard_manager or WildcardManager()
        self._context = SamplingContext(
            wildcard_manager=wildcard_manager,
            default_sampling_method=SamplingMethod.COMBINATORIAL,
            ignore_whitespace=ignore_whitespace,
            parser_config=parser_config,
        )

    def generate(  # type: ignore[override]
        self,
        template: str | None,
        max_prompts: int | None = constants.MAX_IMAGES,
    ) -> list[str]:
        return list(self._context.sample_prompts((template or ""), max_prompts))



from __future__ import annotations

import re
import warnings
from typing import cast

import pyparsing as pp

from dynamicprompts.commands import (
    Command,
    LiteralCommand,
    SamplingMethod,
    SequenceCommand,
    VariantCommand,
    VariantOption,
    WildcardCommand,
)
from dynamicprompts.parser.action_builder import ActionBuilder
from dynamicprompts.parser.config import ParserConfig, default_parser_config

real_num1 = pp.Combine(pp.Word(pp.nums) + "." + pp.Word(pp.nums))
real_num2 = pp.Combine(pp.Word(pp.nums) + ".")
real_num3 = pp.Combine("." + pp.Word(pp.nums))
real_num4 = pp.Word(pp.nums)

real_num = real_num1 | real_num2 | real_num3 | real_num4
sampler_random = pp.Char("~")
sampler_combinatorial = pp.Char("!")
sampler_cyclical = pp.Char("@")
sampler_symbol = sampler_random | sampler_combinatorial | sampler_cyclical

sampler_symbol_to_method = {
    "~": SamplingMethod.RANDOM,
    "!": SamplingMethod.COMBINATORIAL,
    "@": SamplingMethod.CYCLICAL,
}


class Parser:
    """
    A parser for a prompt grammar which is roughly as follows:

    <prompt> ::= (<chunk>)*
    <variant_prompt> ::= (<variant_chunk>)*
    <chunk> ::= <variants> | <wildcard> | <literal_sequence>
    <variant_chunk> ::= <variants> | <wildcard> | <variant_literal_sequence>
    <variants> ::= <variant_start> <sampling_method>?(<bound><separator>?)? <variants_list>? <variant_end>
    <variants_list> ::= <variant> ("|" <variant>)*
    <variant> ::= <weight>? <variant_prompt>
    <weight> ::= <real> | <integer>
    <variant_start> ::= "{"  # Can be configured to an arbitrary string
    <variant_end> ::= "}"    # Can be configured to an arbitrary string
    <sampling_method> ::= "!"|"~"|"@"
    <bound> :: <integer>(-<integer)?$$
    <separator> ::= [^$}]+$$
    <wildcard> ::= <wildcard_enclosure> <sampling_method> <path> <wildcard_enclosure>
    <wildcard_enclosure> ::= "__" # Can be configured to an arbitrary string
    <path>::=  ~"__" + [^{}#]+"
    <literal>:=[^#<variant_start>]+
    <variant_literal>:=[^#$|<variant_start><variant_end>]+
    <literal_sequence> ::= <literal>+
    <variant_literal_sequence> ::= <variant_literal>+

    Note that whitespace is preserved in case it is significant to the user.
    """

    def __init__(self, builder: ActionBuilder):
        warnings.warn(
            f"{self.__class__.__qualname__} is deprecated and will be removed in a future version. "
            "Instead, directly call `parse(prompt)`.",
            DeprecationWarning,
        )

        self._builder = builder
        self._prompt = create_parser(parser_config=default_parser_config)

    @property
    def prompt(self):
        return self._prompt

    def parse(self, prompt: str) -> SequenceCommand:
        tokens = self.prompt.parse_string(prompt, parse_all=True)
        return cast(SequenceCommand, tokens[0])


def _configure_range() -> pp.ParserElement:
    hyphen = pp.Suppress("-")
    variant_delim = pp.Suppress("$$")

    # Exclude:
    # - $, which is used to indicate the end of the separator definition i.e. {1$$ and $$X|Y|Z}
    # - }, which is used to indicate the end of a variant
    # Allowed:
    # - | is allowed as a separator
    separator = pp.Word(pp.printables + " ", exclude_chars="$}")(
        "separator",
    ).leave_whitespace()

    bound = pp.common.integer
    bound_range1 = bound("exact")
    bound_range2 = bound("lower") + hyphen
    bound_range3 = hyphen + bound("upper")
    bound_range4 = bound("lower") + hyphen + bound("upper")

    bound_range = pp.Group(
        bound_range4 | bound_range3 | bound_range2 | bound_range1,
    )
    bound_expr = pp.Group(
        bound_range("range")
        + variant_delim
        + pp.Opt(separator + variant_delim, default=",")("separator"),
    )

    return bound_expr


def _configure_wildcard(
    parser_config: ParserConfig,
) -> pp.ParserElement:
    wildcard_path = pp.Regex(
        r"((?!" + re.escape(parser_config.wildcard_wrap) + r")[^{}#])+",
    )("path").leave_whitespace()
    wildcard_enclosure = pp.Suppress(parser_config.wildcard_wrap)
    wildcard = (
        wildcard_enclosure
        + pp.Optional(sampler_symbol)("sampling_method")
        + wildcard_path
        + wildcard_enclosure
    )

    return wildcard("wildcard").leave_whitespace()


def _configure_literal_sequence(
    parser_config: ParserConfig,
    is_variant_literal: bool = False,
) -> pp.ParserElement:
    # Characters that are not allowed in a literal
    # - { denotes the start of a variant (or whatever variant_start is set to  )
    # - # denotes the start of a comment
    non_literal_chars = rf"#{parser_config.variant_start}"

    if is_variant_literal:
        # Inside a variant the following characters are also not allowed
        # - } denotes the end of a variant (or whatever right brace is set to)
        # - | denotes the end of a variant option
        # - $ denotes the end of a bound expression
        non_literal_chars += rf"|${parser_config.variant_end}"

    non_literal_chars = re.escape(non_literal_chars)
    literal = pp.Regex(
        rf"((?!{re.escape(parser_config.wildcard_wrap)})[^{non_literal_chars}])+",
    )(
        "literal",
    ).leave_whitespace()
    literal_sequence = pp.OneOrMore(literal)

    return literal_sequence


def _create_weight_parser() -> pp.ParserElement:
    weight_delim = pp.Suppress("::")
    weight = (pp.common.real | pp.common.integer) + weight_delim

    return weight


def _configure_variants(
    bound_expr: pp.ParserElement,
    prompt: pp.ParserElement,
    *,
    parser_config: ParserConfig,
) -> pp.ParserElement:
    weight = _create_weight_parser()
    variant_start = pp.Suppress(parser_config.variant_start)
    variant_end = pp.Suppress(parser_config.variant_end)

    variant = pp.Group(
        pp.Optional(pp.White())
        + pp.Opt(weight, default=1)("weight")
        + prompt()("val")
        + pp.Optional(pp.White()),
    )
    variants_list = pp.Group(pp.delimited_list(variant, delim="|"))

    variants = pp.Group(
        variant_start
        + pp.Optional(pp.White())
        + pp.Optional(sampler_symbol)("sampling_method")
        + pp.Opt(bound_expr)("bound_expr")
        + pp.Optional(pp.White())
        + variants_list("variants")
        + pp.Optional(pp.White())
        + variant_end,
    )

    return variants.leave_whitespace()


def _parse_literal_command(parse_result: pp.ParseResults) -> LiteralCommand:
    s = " ".join(parse_result)
    return LiteralCommand(s)


def _parse_sequence_or_single_command(parse_result: pp.ParseResults) -> Command:
    children = list(parse_result)
    assert all(isinstance(c, Command) for c in children)
    if len(children) == 1:  # If there is only one child, return it directly
        return children[0]
    return SequenceCommand(tokens=children)


def _parse_variant_command(parse_result: pp.ParseResults) -> VariantCommand:
    assert len(parse_result) == 1
    parts = parse_result[0].as_dict()

    sampling_method_symbol = parts.get("sampling_method")
    sampling_method = _parse_sampling_method(sampling_method_symbol)

    variants = [
        VariantOption(value=v["val"], weight=float(v["weight"][0]))
        for v in parts["variants"]
    ]
    if "bound_expr" in parts:
        min_bound, max_bound, separator = _parse_bound_expr(
            parts["bound_expr"],
            max_options=len(variants),
        )
    else:
        min_bound = max_bound = 1
        separator = ","
    return VariantCommand(
        variants,
        min_bound=min_bound,
        max_bound=max_bound,
        separator=separator,
        sampling_method=sampling_method,
    )


def _parse_sampling_method(sampling_method_symbol: str | None) -> SamplingMethod | None:
    if sampling_method_symbol is None:
        return None
    try:
        return sampler_symbol_to_method[sampling_method_symbol]
    except KeyError:
        raise ValueError(
            f"Unexpected sampling method: {sampling_method_symbol}.",
        ) from None


def _parse_wildcard_command(parse_result: pp.ParseResults) -> WildcardCommand:
    parts = parse_result.as_dict()
    wildcard = parts.get("path")

    sampling_method_symbol = parts.get("sampling_method")
    sampling_method = _parse_sampling_method(sampling_method_symbol)

    assert isinstance(wildcard, str)
    return WildcardCommand(wildcard=wildcard, sampling_method=sampling_method)


def _parse_bound_expr(expr, max_options):
    lbound = 1
    ubound = max_options
    separator = ","

    expr = expr[0]

    if "range" in expr:
        rng = expr["range"]
        if "exact" in rng:
            lbound = ubound = rng["exact"]
        else:
            if "lower" in expr["range"]:
                lbound = int(expr["range"]["lower"])
            if "upper" in expr["range"]:
                ubound = int(expr["range"]["upper"])

    if "separator" in expr:
        separator = expr["separator"][0]

    return lbound, ubound, separator


def create_parser(
    *,
    parser_config: ParserConfig,
) -> pp.ParserElement:
    bound_expr = _configure_range()

    prompt = pp.Forward()
    variant_prompt = pp.Forward()

    wildcard = _configure_wildcard(parser_config=parser_config)
    literal_sequence = _configure_literal_sequence(parser_config=parser_config)
    variant_literal_sequence = _configure_literal_sequence(
        is_variant_literal=True,
        parser_config=parser_config,
    )
    variants = _configure_variants(
        bound_expr,
        variant_prompt,
        parser_config=parser_config,
    )

    chunk = variants | wildcard | literal_sequence
    variant_chunk = variants | wildcard | variant_literal_sequence

    prompt <<= pp.ZeroOrMore(chunk)("prompt")
    variant_prompt <<= pp.ZeroOrMore(variant_chunk)("prompt")

    # Configure comments
    prompt.ignore("#" + pp.restOfLine)
    prompt.ignore("//" + pp.restOfLine)
    prompt.ignore(pp.c_style_comment)

    wildcard.set_parse_action(_parse_wildcard_command)
    variants.set_parse_action(_parse_variant_command)
    literal_sequence.set_parse_action(_parse_literal_command)
    variant_literal_sequence.set_parse_action(_parse_literal_command)
    prompt.set_parse_action(_parse_sequence_or_single_command)
    variant_prompt.set_parse_action(_parse_sequence_or_single_command)
    return prompt


def parse(
    prompt: str,
    parser_config: ParserConfig = default_parser_config,
) -> Command:
    """
    Parse a prompt string into a commands.
    :param prompt: The prompt string to parse.
    :return: A command representing the parsed prompt.
    """

    tokens = create_parser(parser_config=parser_config).parse_string(
        prompt,
        parse_all=True,
    )
    assert (
        tokens and len(tokens) == 1
    )  # If we have more than one token, the prompt is invalid...
    tok = tokens[0]

    assert isinstance(tok, Command)
    return tok



from __future__ import annotations

import dataclasses
import logging
from typing import Generator, Iterable

from dynamicprompts.commands.base import Command
from dynamicprompts.commands.literal_command import LiteralCommand
from dynamicprompts.enums import SamplingMethod

logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True)
class VariantOption:
    value: Command
    weight: float = 1.0


@dataclasses.dataclass(frozen=True)
class VariantCommand(Command):
    variants: list[VariantOption]
    min_bound: int = 1
    max_bound: int = 1
    separator: str = ","
    sampling_method: SamplingMethod | None = None

    def __post_init__(self):
        min_bound, max_bound = sorted((self.min_bound, self.max_bound))
        min_bound = max(0, min_bound)
        object.__setattr__(self, "min_bound", min_bound)
        object.__setattr__(self, "max_bound", max_bound)

    def __len__(self) -> int:
        return len(self.variants)

    def __getitem__(self, index: int) -> VariantOption:
        return self.variants[index]

    def __iter__(self) -> Iterable[VariantOption]:
        return iter(self.variants)

    @property
    def weights(self) -> list[float]:
        return [p.weight for p in self.variants]

    @property
    def values(self) -> list[Command]:
        return [p.value for p in self.variants]

    def adjust_range(self) -> VariantCommand:
        min_bound = min(self.min_bound, len(self.values))
        max_bound = min(self.max_bound, len(self.values))
        return dataclasses.replace(self, min_bound=min_bound, max_bound=max_bound)

    @classmethod
    def from_literals_and_weights(
        cls,
        literals: list[str],
        weights: list[float] | None = None,
        min_bound: int = 1,
        max_bound: int = 1,
        separator: str = ",",
        sampling_method: SamplingMethod | None = None,
    ) -> VariantCommand:
        vals = [LiteralCommand(str(v)) for v in literals]
        if weights is None:
            weights = [1.0] * len(vals)
        assert len(vals) == len(weights), "Must have same number of weights as values"
        return VariantCommand(
            variants=[VariantOption(v, w) for v, w in zip(vals, weights)],
            min_bound=min_bound,
            max_bound=max_bound,
            separator=separator,
            sampling_method=sampling_method,
        )

    def get_value_combinations(
        self,
        k: int,
        values=None,
    ) -> Generator[list[Command], None, None]:
        if values is None:
            values = self.values

        if k <= 0:
            yield []
        else:
            for value in values:
                other_values = [v for v in values if v != value]
                for item in self.get_value_combinations(k - 1, values=other_values):
                    yield [value] + item



from __future__ import annotations

from typing import Iterable

from dynamicprompts.commands import Command
from dynamicprompts.sampling_context import SamplingContext
from dynamicprompts.types import CommandList, StringGen


class CommandCollection:
    """
    A class that holds a collection of commands and manages generating values for them.
    """

    def __init__(self, commands: Iterable[Command], context: SamplingContext) -> None:
        self._commands = list(commands)
        self._generators = [context.generator_from_command(c) for c in self._commands]
        self._values: list[str | None] = [next(g) for g in self._generators]

    def get_value(self, command: Command) -> str | None:
        try:
            index = self._commands.index(command)
        except ValueError:
            raise ValueError(f"Command {command} not in collection") from None

        generator = self._generators[index]
        value = self._values[index]

        try:
            self._values[index] = next(generator)
        except StopIteration:
            self._values[index] = None

        return value

    @property
    def commands(self) -> CommandList:
        return self._commands

    @property
    def generators(self) -> list[StringGen]:
        return self._generators



from __future__ import annotations

import random
from collections import OrderedDict
from itertools import cycle
from typing import Iterable, TypeVar

from dynamicprompts.types import StringGen

T = TypeVar("T")


def removeprefix(s: str, prefix: str) -> str:
    return s[len(prefix) :] if s.startswith(prefix) else s


def removesuffix(s: str, suffix: str) -> str:
    return s[: -len(suffix)] if s.endswith(suffix) else s


def squash_whitespace(s: str) -> str:
    return " ".join(s.split())


def is_empty_line(line: str | None) -> bool:
    return line is None or line.strip() == "" or line.strip().startswith("#")


def dedupe(arr: list[str]) -> tuple[str, ...]:
    ordered_dict = OrderedDict.fromkeys(arr)
    return tuple(ordered_dict.keys())


def rotate_all(generators: Iterable[StringGen]) -> list[str]:
    return [next(gen) for gen in generators]


def rotate_and_join(
    generators: Iterable[StringGen],
    *,
    separator: str,
) -> str:
    return separator.join(rotate_all(generators))


def next_sampler_next_value(
    samplers: Iterable[StringGen],
) -> StringGen:
    yield from (next(iter(sampler)) for sampler in cycle(samplers))


def cross(l1: Iterable[T], l2: Iterable[T]) -> Iterable[tuple[T, T]]:
    for e1 in l1:
        for e2 in l2:
            if e1 != e2:
                yield (e1, e2)


def choose_without_replacement(
    values: list[T],
    *,
    weights: list[float],
    num_choices: int,
    rand=random,
) -> list[T]:
    values = values.copy()
    weights = weights.copy()

    if len(values) < num_choices:
        raise ValueError(
            f"You asked for {num_choices} values, but only {len(values)} are available in {values}.",
        )

    if not weights:
        weights = [1.0 for _ in range(len(values))]

    if len(values) != len(weights):
        raise ValueError(
            f"Number of values ({len(values)}) and weights ({len(weights)}) must be the same.",
        )

    if len(values) == 0:
        return []
    elif len(values) == 1:
        return values
    else:
        chosen_values = []
        for _ in range(num_choices):
            chosen = rand.choices(values, weights=weights, k=1)[0]
            chosen_values.append(chosen)
            weights.remove(weights[values.index(chosen)])
            values.remove(chosen)
        return chosen_values



from __future__ import annotations

import logging
from random import Random

from dynamicprompts.commands.base import SamplingMethod
from dynamicprompts.constants import DEFAULT_RANDOM
from dynamicprompts.generators.promptgenerator import PromptGenerator
from dynamicprompts.parser.config import ParserConfig, default_parser_config
from dynamicprompts.sampling_context import SamplingContext
from dynamicprompts.wildcardmanager import WildcardManager

logger = logging.getLogger(__name__)


def _get_random(*, seed: int | None, unlink_seed_from_prompt: bool) -> Random:
    if unlink_seed_from_prompt:
        return DEFAULT_RANDOM
    rand = Random()
    if seed is not None:
        rand.seed(seed)
    return rand


class RandomPromptGenerator(PromptGenerator):
    def __init__(
        self,
        wildcard_manager: WildcardManager | None = None,
        seed: int | None = None,
        unlink_seed_from_prompt: bool = False,
        ignore_whitespace: bool = False,
        parser_config: ParserConfig = default_parser_config,
    ) -> None:
        wildcard_manager = wildcard_manager or WildcardManager()
        self._context = SamplingContext(
            wildcard_manager=wildcard_manager,
            default_sampling_method=SamplingMethod.RANDOM,
            ignore_whitespace=ignore_whitespace,
            parser_config=parser_config,
            rand=_get_random(
                seed=seed,
                unlink_seed_from_prompt=unlink_seed_from_prompt,
            ),
        )

    def generate(self, template: str | None, num_images: int = 1) -> list[str]:
        return list(self._context.sample_prompts((template or ""), num_images))



from dynamicprompts.commands import (
    VariantCommand,
    WildcardCommand,
)
from dynamicprompts.sampling_context import SamplingContext


def wildcard_to_variant(
    wildcard: WildcardCommand,
    *,
    context: SamplingContext,
    min_bound=1,
    max_bound=1,
    separator=",",
) -> VariantCommand:
    wm = context.wildcard_manager
    wildcard_values = wm.get_all_values(wildcard.wildcard)
    min_bound = min(min_bound, len(wildcard_values))
    max_bound = min(max_bound, len(wildcard_values))

    wildcard_variant = VariantCommand.from_literals_and_weights(
        wildcard_values,
        min_bound=min_bound,
        max_bound=max_bound,
        separator=separator,
        sampling_method=wildcard.sampling_method,
    )
    return wildcard_variant



from __future__ import annotations

import dataclasses

from dynamicprompts.commands import Command
from dynamicprompts.enums import SamplingMethod


@dataclasses.dataclass(frozen=True)
class LiteralCommand(Command):
    literal: str
    sampling_method: SamplingMethod | None = None

    def __add__(self, other):
        if isinstance(other, LiteralCommand):
            return LiteralCommand(f"{self.literal} {other.literal}")
        raise TypeError(f"Cannot concatenate LiteralCommand with {other}")



from dynamicprompts.samplers.base import Sampler
from dynamicprompts.samplers.combinatorial import CombinatorialSampler
from dynamicprompts.samplers.cycle import CyclicalSampler
from dynamicprompts.samplers.random import RandomSampler

__all__ = [
    "CombinatorialSampler",
    "RandomSampler",
    "Sampler",
    "CyclicalSampler",
]



from __future__ import annotations

import logging
from itertools import cycle
from typing import Generator, Iterable, cast

from dynamicprompts.commands import (
    Command,
    SamplingMethod,
    VariantCommand,
    WildcardCommand,
)
from dynamicprompts.samplers.base import Sampler
from dynamicprompts.samplers.utils import wildcard_to_variant
from dynamicprompts.sampling_context import SamplingContext
from dynamicprompts.types import StringGen, to_string_gen
from dynamicprompts.utils import next_sampler_next_value

logger = logging.getLogger(__name__)


def get_arrs(gen: StringGen, rest_gen: Iterable[list[str]]):
    try:
        for r in rest_gen:
            yield [next(gen)] + r
    except StopIteration:
        yield []


def _get_combination_samples(
    combo: list[Command],
    sampling_context: SamplingContext,
) -> Generator[list[str], None, None]:
    try:
        if len(combo) == 0:
            while True:
                yield []
        else:
            c_1, c_rest = combo[0], combo[1:]

            gen = sampling_context.generator_from_command(c_1)
            rest_gen = _get_combination_samples(c_rest, sampling_context)
            arrs = get_arrs(gen, rest_gen)

            yield from arrs
    except StopIteration:
        return


class CyclicalSampler(Sampler):
    def _get_variant(
        self,
        command: VariantCommand,
        sampling_context: SamplingContext,
    ) -> StringGen:
        is_wildcard_variant = len(command.values) == 1 and isinstance(
            command.values[0],
            WildcardCommand,
        )

        if len(command.values) == 0:
            return
        elif is_wildcard_variant:
            wildcard_command = cast(WildcardCommand, command.values[0])
            wildcard_variant = wildcard_to_variant(
                wildcard_command,
                context=sampling_context,
                min_bound=command.min_bound,
                max_bound=command.max_bound,
                separator=command.separator,
            )
            yield from self._get_variant(wildcard_variant, sampling_context)
        else:
            command = command.adjust_range()

            combinations = (
                combo
                for bound in range(command.min_bound, command.max_bound + 1)
                for combo in command.get_value_combinations(bound)
            )

            combination_samplers = (
                (
                    command.separator.join(sample)
                    for sample in _get_combination_samples(combo, sampling_context)
                )
                for combo in combinations
            )

            while True:
                yield from next_sampler_next_value(cycle(combination_samplers))

    def _get_wildcard(
        self,
        command: WildcardCommand,
        sampling_context: SamplingContext,
    ):
        values = sampling_context.wildcard_manager.get_all_values(command.wildcard)
        new_context = sampling_context.with_sampling_method(SamplingMethod.CYCLICAL)
        value_samplers = [new_context.sample_prompts(val) for val in values]
        value_string_gens = [to_string_gen(val) for val in value_samplers]

        if len(values) == 0:
            logger.warning(f"No values found for wildcard {command.wildcard}")
            ww = sampling_context.parser_config.wildcard_wrap
            while True:
                yield f"{ww}{command.wildcard}{ww}"
        else:
            yield from next_sampler_next_value(value_string_gens)



from __future__ import annotations

import logging
import random
from itertools import permutations
from typing import Any, Callable, Iterable, cast

from jinja2 import Environment, pass_environment
from jinja2.ext import Extension
from jinja2.nodes import CallBlock
from jinja2.parser import Parser

logger = logging.getLogger(__name__)


def choice(*items) -> Any:
    return random.choice(items)


def weighted_choice(*items) -> Any:
    population, weights = zip(*items)
    return random.choices(population, weights=weights)[0]


def permutation(
    items: Iterable[Any],
    low: int,
    high: int | None = None,
) -> list[Any]:
    vars: list[Any] = []
    if high is None:
        high = low

    for i in range(low, high + 1):
        vars.extend(permutations(items, i))

    return vars


@pass_environment
def wildcard(environment: Environment, wildcard_name: str) -> list[str]:
    from dynamicprompts.generators import CombinatorialPromptGenerator
    from dynamicprompts.wildcardmanager import WildcardManager

    wm: WildcardManager = environment.globals["wildcard_manager"]  # type: ignore
    generator: CombinatorialPromptGenerator = environment.globals["generators"]["combinatorial"]  # type: ignore

    return list(generator.generate(wm.to_wildcard(wildcard_name)))
    return combinatorial_sample(environment, wildcard_name)


@pass_environment
def random_sample(environment: Environment, prompt: str) -> str:
    generator = environment.globals["generators"]["random"]  # type: ignore

    return list(generator.generate(prompt))[0]


@pass_environment
def combinatorial_sample(environment: Environment, prompt: str) -> list[str]:
    generator = environment.globals["generators"]["combinatorial"]  # type: ignore

    return list(generator.generate(prompt))


class PromptExtension(Extension):
    tags = {"prompt"}

    def __init__(self, environment: Environment) -> None:
        super().__init__(environment)
        environment.extend(prompt_blocks=[])

    def parse(self, parser: Parser) -> CallBlock:
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(
            ("name:endprompt",),
            drop_needle=True,
        )
        node = CallBlock(
            self.call_method("_prompt", []),
            [],
            [],
            body,
        )
        node.set_lineno(lineno)
        return node

    def _prompt(self, caller: Callable) -> Any:
        value = caller()
        prompt_blocks = cast(list, self.environment.globals["prompt_blocks"])
        prompt_blocks.append(value)
        return value


DYNAMICPROMPTS_FUNCTIONS: dict[str, Any] = {
    "choice": choice,
    "weighted_choice": weighted_choice,
    "random": random.random,
    "randint": random.randint,
    "permutations": permutation,
    "wildcard": wildcard,
    "random_sample": random_sample,
    "all_combinations": combinatorial_sample,
}



from __future__ import annotations

import dataclasses

from dynamicprompts.commands.base import Command
from dynamicprompts.enums import SamplingMethod


@dataclasses.dataclass(frozen=True)
class WildcardCommand(Command):
    wildcard: str
    sampling_method: SamplingMethod | None = None

    def __post_init__(self):
        if not isinstance(self.wildcard, str):
            raise TypeError(f"Wildcard must be a string, not {type(self.wildcard)}")



from __future__ import annotations

import dataclasses
import warnings
from itertools import islice
from random import Random
from typing import TYPE_CHECKING, Iterable

from dynamicprompts.commands import Command
from dynamicprompts.constants import DEFAULT_RANDOM
from dynamicprompts.enums import SamplingMethod
from dynamicprompts.parser.config import ParserConfig, default_parser_config
from dynamicprompts.types import StringGen
from dynamicprompts.utils import squash_whitespace
from dynamicprompts.wildcardmanager import WildcardManager

if TYPE_CHECKING:
    from dynamicprompts.samplers import Sampler


def _build_default_samplers():
    from dynamicprompts.samplers.combinatorial import CombinatorialSampler
    from dynamicprompts.samplers.cycle import CyclicalSampler
    from dynamicprompts.samplers.random import RandomSampler

    return {
        SamplingMethod.COMBINATORIAL: CombinatorialSampler(),
        SamplingMethod.CYCLICAL: CyclicalSampler(),
        SamplingMethod.RANDOM: RandomSampler(),
    }


@dataclasses.dataclass(frozen=True)
class SamplingContext:
    default_sampling_method: SamplingMethod
    wildcard_manager: WildcardManager
    samplers: dict[SamplingMethod, Sampler] = dataclasses.field(
        default_factory=_build_default_samplers,
    )
    ignore_whitespace: bool = False
    parser_config: ParserConfig = default_parser_config
    rand: Random = DEFAULT_RANDOM

    def with_sampling_method(self, sampling_method: SamplingMethod) -> SamplingContext:
        return dataclasses.replace(self, default_sampling_method=sampling_method)

    @property
    def default_sampler(self) -> Sampler:
        return self.samplers[self.default_sampling_method]

    def get_sampler_and_context(
        self,
        command: Command,
    ) -> tuple[Sampler, SamplingContext]:
        """
        Get the correct sampler instance and a sub-context (if necessary) for the given command.
        """
        new_sampling_method = command.sampling_method or self.default_sampling_method
        if new_sampling_method != self.default_sampling_method:
            if self.default_sampling_method.is_nonfinite():  # Within non-finite context
                if not new_sampling_method.is_nonfinite():
                    # ...but using finite method?
                    warnings.warn(
                        f"Command {command} has finite sampling method {new_sampling_method} "
                        f"that can't be nested within this non-finite context {self.default_sampling_method}, "
                        f"so using {self.default_sampling_method} instead.",
                    )
                    new_sampling_method = self.default_sampling_method
            sampler = self.samplers[new_sampling_method]
            context = self.with_sampling_method(new_sampling_method)
        else:
            sampler = self.samplers[self.default_sampling_method]
            context = self
        return sampler, context

    def generator_from_command(self, command: Command) -> StringGen:
        samp, ctx = self.get_sampler_and_context(command)
        return samp.generator_from_command(command, ctx)

    def sample_prompts(
        self,
        prompt: str | Command,
        num_prompts: int | None = None,
    ) -> Iterable[str]:
        """
        Generate prompts from a prompt template.

        :param prompt: The prompt template to generate prompts from.
        :param num_prompts: How many prompts to generate (at most). If None, generate all possible prompts.
        """
        if not prompt:
            return []
        command: Command
        if isinstance(prompt, str):
            from dynamicprompts.parser.parse import parse

            command = parse(prompt, parser_config=self.parser_config)
        elif isinstance(prompt, Command):
            command = prompt
        else:
            raise TypeError(f"Expected prompt to be str or Command, got {type(prompt)}")

        gen = self.generator_from_command(command)

        if self.ignore_whitespace:
            gen = (squash_whitespace(p) for p in gen)

        if num_prompts is None:
            return gen
        return islice(gen, num_prompts)

    def get_effective_sampling_method(self, command: Command) -> SamplingMethod:
        if command.sampling_method:
            return command.sampling_method
        return self.default_sampling_method



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Any

from dynamicprompts import constants
from dynamicprompts.parser.config import default_parser_config
from dynamicprompts.utils import removeprefix, removesuffix
from dynamicprompts.wildcardfile import WildcardFile

logger = logging.getLogger(__name__)


def _is_relative_to(p1: Path, p2: Path) -> bool:
    # Port of Path.is_relative_to from Python 3.9
    try:
        p1.relative_to(p2)
        return True
    except ValueError:
        return False


def _clean_wildcard(wildcard: str, *, wildcard_wrap: str) -> str:
    wildcard = (
        wildcard.replace("/", os.sep)
        .replace("\\", os.sep)  # normalize path separators
        .rstrip(os.sep)  # remove trailing path separator (likely a typo)
    )
    wildcard = removeprefix(wildcard, wildcard_wrap)
    wildcard = removesuffix(wildcard, wildcard_wrap)

    if wildcard.startswith(os.sep):
        raise ValueError(f"Wildcard {wildcard} cannot start with {os.sep}")
    if ".." in wildcard:
        raise ValueError(f"Wildcard can not contain '..': {wildcard}")
    return wildcard


class WildcardManager:
    def __init__(
        self,
        path: Path | None = None,
        wildcard_wrap=default_parser_config.wildcard_wrap,
    ) -> None:
        self._path = path
        self._wildcard_wrap = wildcard_wrap

    @property
    def path(self) -> Path | None:
        """
        The root path of the wildcard manager, if set.
        """
        return self._path

    @property
    def wildcard_wrap(self) -> str:
        """
        The string that is used as the prefix and suffix of a wildcard
        the default is "__" (two underscores)
        """
        return self._wildcard_wrap

    def _directory_exists(self) -> bool:
        return bool(self._path and self._path.is_dir())

    def ensure_directory(self) -> None:
        if not self._path:
            return
        try:
            self._path.mkdir(parents=True, exist_ok=True)
        except Exception:
            logger.exception(f"Failed to create directory {self._path}")

    def get_files(self, relative: bool = False) -> list[Path]:
        if not (self._path and self._directory_exists()):
            return []

        files = list(self._path.rglob(f"*.{constants.WILDCARD_SUFFIX}"))
        if relative:
            files = [f.relative_to(self._path) for f in files]

        return list(files)

    def match_files(self, wildcard: str) -> list[WildcardFile]:
        if not self._path:
            return []
        try:
            wildcard = _clean_wildcard(wildcard, wildcard_wrap=self._wildcard_wrap)
        except ValueError:
            logger.warning(f"Invalid wildcard: {wildcard}", exc_info=True)
            return []

        return [
            WildcardFile(path, name=self.path_to_wildcard_without_separators(path))
            for path in self._path.rglob(f"{wildcard}.{constants.WILDCARD_SUFFIX}")
            if _is_relative_to(path.absolute(), self._path)
        ]

    def wildcard_to_path(self, wildcard: str) -> Path:
        if not self._path:  # pragma: no cover
            raise ValueError("Can't call wildcard_to_path without a path set")
        return (
            self._path / _clean_wildcard(wildcard, wildcard_wrap=self._wildcard_wrap)
        ).with_suffix(
            f".{constants.WILDCARD_SUFFIX}",
        )

    def path_to_wildcard_without_separators(self, path: Path) -> str:
        if not self._path:  # pragma: no cover
            raise ValueError(
                "Can't call path_to_wildcard_without_separators without a path set",
            )
        rel_path = path.relative_to(self._path)
        return str(rel_path.with_suffix("")).replace(os.sep, "/")

    def path_to_wildcard(self, path: Path) -> str:
        return f"{self._wildcard_wrap}{self.path_to_wildcard_without_separators(path)}{self._wildcard_wrap}"

    def get_wildcards(self) -> list[str]:
        files = self.get_files(relative=True)
        return [self.path_to_wildcard(f) for f in files]

    def get_all_values(self, wildcard: str) -> list[str]:
        files = self.match_files(wildcard)
        return sorted(set().union(*[f.get_wildcards() for f in files]))

    def to_wildcard(self, name: str) -> str:
        """
        Wrap `name` in the wildcard wrap string if it is not already wrapped.
        """
        ww = self._wildcard_wrap
        if not name.startswith(ww):
            name = f"{ww}{name}"
        if not name.endswith(ww):
            name = f"{name}{ww}"
        return name

    # TODO: the return type is actually a recursive type (replace that Any)
    def get_wildcard_hierarchy(
        self,
        path: Path | None = None,
    ) -> tuple[list[str], dict[str, Any]]:
        if path is None:
            path = self._path
        if not path:
            return ([], {})

        files = path.glob(f"*.{constants.WILDCARD_SUFFIX}")
        wildcards = sorted(self.path_to_wildcard(f) for f in files)
        directories = sorted(d for d in path.iterdir() if d.is_dir())

        hierarchy = {d.name: self.get_wildcard_hierarchy(d) for d in directories}
        return (wildcards, hierarchy)

    def is_wildcard(self, text: str) -> bool:
        return text.startswith(self.wildcard_wrap) and text.endswith(self.wildcard_wrap)

    def get_collection_path(self) -> Path:
        if not self._path:  # pragma: no cover
            raise ValueError("Can't call get_collection_path without a path set")
        return self._path.parent / "collections"

    def get_collections(self) -> list[str]:
        return list(self.get_collection_dirs().keys())

    def get_collection_dirs(self) -> dict[str, Path]:
        collection_path = self.get_collection_path()
        collection_dirs = [x for x in collection_path.iterdir() if x.is_dir()]
        collection_names = [
            str(c.relative_to(collection_path)) for c in collection_dirs
        ]

        return dict(zip(collection_names, collection_dirs))



from dynamicprompts.generators.batched_combinatorial import (
    BatchedCombinatorialPromptGenerator,
)
from dynamicprompts.generators.combinatorial import CombinatorialPromptGenerator
from dynamicprompts.generators.dummygenerator import DummyGenerator
from dynamicprompts.generators.feelinglucky import FeelingLuckyGenerator
from dynamicprompts.generators.jinjagenerator import JinjaGenerator
from dynamicprompts.generators.promptgenerator import PromptGenerator
from dynamicprompts.generators.randomprompt import RandomPromptGenerator

__all__ = [
    "BatchedCombinatorialPromptGenerator",
    "CombinatorialPromptGenerator",
    "DummyGenerator",
    "FeelingLuckyGenerator",
    "JinjaGenerator",
    "PromptGenerator",
    "RandomPromptGenerator",
]



from __future__ import annotations

import logging

from jinja2 import Environment
from jinja2.exceptions import TemplateSyntaxError

from dynamicprompts.generators.combinatorial import CombinatorialPromptGenerator
from dynamicprompts.generators.promptgenerator import (
    GeneratorException,
    PromptGenerator,
)
from dynamicprompts.generators.randomprompt import RandomPromptGenerator
from dynamicprompts.jinja_extensions import DYNAMICPROMPTS_FUNCTIONS, PromptExtension
from dynamicprompts.parser.config import ParserConfig, default_parser_config
from dynamicprompts.utils import squash_whitespace
from dynamicprompts.wildcardmanager import WildcardManager

logger = logging.getLogger(__name__)


class JinjaGenerator(PromptGenerator):
    def __init__(
        self,
        wildcard_manager: WildcardManager | None = None,
        context: dict | None = None,
        parser_config: ParserConfig = default_parser_config,
        unlink_seed_from_prompt: bool = False,
        ignore_whitespace: bool = False,
        limit_prompts=False,
    ) -> None:
        """
        Initialize a JinjaGenerator

        :param wildcard_manager: The wildcard manager to use for this generator, if None a null manager will be used
        :param context: The context to use for this generator, values in this dict will be available to the template
        :param parser_config: The parser config to use for this generator
        :param unlink_seed_from_prompt: Passed to the RandomPromptGenerator to allow random prompts even if the seed is fixed
        :param ignore_whitespace: Passed to the RandomPromptGenerator to ignore whitespace when generating prompts
        :param limit_prompts: Whether to limit the number of prompts generated to num_prompts, default is to generate num_prompts * num_prompts_in_template

        """

        self._wildcard_manager = wildcard_manager or WildcardManager()
        self._parser_config = parser_config
        self._unlink_seed_from_prompt = unlink_seed_from_prompt
        self._generators = {
            "random": RandomPromptGenerator(
                self._wildcard_manager,
                parser_config=self._parser_config,
                unlink_seed_from_prompt=self._unlink_seed_from_prompt,
                ignore_whitespace=ignore_whitespace,
            ),
            "combinatorial": CombinatorialPromptGenerator(
                self._wildcard_manager,
                parser_config=self._parser_config,
                ignore_whitespace=ignore_whitespace,
            ),
        }

        self._context = context or {}
        self._limit_prompts = limit_prompts
        self._ignore_whitespace = ignore_whitespace

    def generate(self, template: str, num_prompts: int = 1) -> list[str]:
        env = Environment(extensions=[PromptExtension])
        prompt_blocks: list[str] = []
        env.globals.update(
            {
                "wildcard_manager": self._wildcard_manager,
                "prompt_blocks": prompt_blocks,
                "parser_config": self._parser_config,
                "generators": self._generators,
                **DYNAMICPROMPTS_FUNCTIONS,
            },
        )

        try:
            jinja_template = env.from_string(template)
        except TemplateSyntaxError as e:
            logger.exception(e)
            raise GeneratorException(e.message) from e

        prompts = []
        for i in range(num_prompts):
            s = jinja_template.render(**self._context)
            prompts.append(s)

        if prompt_blocks:
            prompts = prompt_blocks

        if self._limit_prompts:
            prompts = prompts[0:num_prompts]

        if self._ignore_whitespace:
            prompts = [squash_whitespace(p) for p in prompts]
        return prompts



from dynamicprompts.commands.base import Command, SamplingMethod
from dynamicprompts.commands.literal_command import LiteralCommand
from dynamicprompts.commands.sequence_command import SequenceCommand
from dynamicprompts.commands.variant_command import VariantCommand, VariantOption
from dynamicprompts.commands.wildcard_command import WildcardCommand

__all__ = [
    "Command",
    "LiteralCommand",
    "SequenceCommand",
    "VariantCommand",
    "VariantOption",
    "WildcardCommand",
    "SamplingMethod",
]



from dataclasses import dataclass


@dataclass(frozen=True)
class ParserConfig:
    variant_start: str = "{"
    variant_end: str = "}"
    wildcard_wrap: str = "__"


default_parser_config = ParserConfig()



from __future__ import annotations

import logging
from random import Random

from dynamicprompts.commands import (
    Command,
    VariantCommand,
    WildcardCommand,
)
from dynamicprompts.samplers.base import Sampler
from dynamicprompts.samplers.utils import wildcard_to_variant
from dynamicprompts.sampling_context import SamplingContext
from dynamicprompts.types import StringGen
from dynamicprompts.utils import choose_without_replacement, rotate_and_join

logger = logging.getLogger(__name__)


class RandomSampler(Sampler):
    def _get_variant_choices(
        self,
        values: list[Command],
        weights: list[float],
        num_choices: int,
        rand: Random,
    ) -> list[Command]:
        # Wraps choose_without_replacement for ease of testing

        return choose_without_replacement(
            values,
            weights=weights,
            num_choices=num_choices,
            rand=rand,
        )

    def _get_variant_num_choices(
        self,
        command: VariantCommand,
        context: SamplingContext,
    ) -> int:
        # Wraps randint for ease of testing
        return context.rand.randint(
            command.min_bound,
            command.max_bound,
        )

    def _get_wildcard_choice(self, context: SamplingContext, values: list[str]) -> str:
        # Wraps choice for ease of testing
        return context.rand.choice(values)

    def _get_variant(
        self,
        command: VariantCommand,
        context: SamplingContext,
    ) -> StringGen:
        if len(command.values) == 0:
            return
        elif len(command.values) == 1:
            if isinstance(command.values[0], WildcardCommand):
                wildcard_variant = wildcard_to_variant(
                    command.values[0],
                    context=context,
                    min_bound=command.min_bound,
                    max_bound=command.max_bound,
                    separator=command.separator,
                )

                yield from self._get_variant(wildcard_variant, context)
            else:
                yield from context.generator_from_command(
                    command.values[0],
                )
            return
        while True:
            command = command.adjust_range()

            num_choices = min(
                command.max_bound,
                self._get_variant_num_choices(command, context),
            )

            selected_commands = self._get_variant_choices(
                command.values,
                weights=command.weights,
                num_choices=num_choices,
                rand=context.rand,
            )
            sub_generators = [
                context.generator_from_command(c) for c in selected_commands
            ]

            if len(sub_generators) == 0:
                yield ""
            else:
                yield rotate_and_join(
                    sub_generators,
                    separator=command.separator,
                )

    def _get_wildcard(
        self,
        command: WildcardCommand,
        context: SamplingContext,
    ) -> StringGen:
        values = context.wildcard_manager.get_all_values(command.wildcard)

        if len(values) == 0:
            logger.warning(f"No values found for wildcard {command.wildcard}")

        while True:
            if len(values) == 0:
                yield context.wildcard_manager.to_wildcard(command.wildcard)
            else:
                value = self._get_wildcard_choice(context, values)
                yield from context.sample_prompts(value, 1)



from __future__ import annotations

import logging
import typing
from typing import cast

from dynamicprompts.commands import (
    Command,
    LiteralCommand,
    SamplingMethod,
    SequenceCommand,
    VariantCommand,
    WildcardCommand,
)
from dynamicprompts.samplers.base import Sampler
from dynamicprompts.samplers.command_collection import CommandCollection
from dynamicprompts.samplers.utils import wildcard_to_variant
from dynamicprompts.sampling_context import SamplingContext
from dynamicprompts.types import StringGen

logger = logging.getLogger(__name__)


def _dedupe(arr: list[str]) -> tuple[str, ...]:
    seen: set[str] = set()
    result: list[str] = []
    for item in arr:
        if item not in seen:
            seen.add(item)
            result.append(item)
    return tuple(result)


def _combo_to_prompt(
    sampling_context: SamplingContext,
    combo: list[Command],
) -> typing.Iterable[list[str]]:
    if len(combo) == 0:
        yield []
        return

    c_1, c_rest = combo[0], combo[1:]
    gen = sampling_context.generator_from_command(c_1)
    if sampling_context.get_effective_sampling_method(c_1).is_nonfinite():
        for rest_prompt in _combo_to_prompt(sampling_context, c_rest):
            val = next(gen)
            yield [val] + rest_prompt
    else:
        for p in gen:
            for rest_prompt in _combo_to_prompt(sampling_context, c_rest):
                if rest_prompt:
                    yield [p] + rest_prompt
                else:
                    yield [p]


class CombinatorialSampler(Sampler):
    def _get_sequence(
        self,
        command: SequenceCommand,
        context: SamplingContext,
    ) -> StringGen:
        non_combo_commands = [
            c
            for c in command.tokens
            if c.sampling_method != SamplingMethod.COMBINATORIAL
        ]
        command_collection = CommandCollection(
            non_combo_commands,
            context=context,
        )
        augmented_tokens = [
            LiteralCommand(
                "",
                sampling_method=SamplingMethod.COMBINATORIAL,
            ),  # sentinel 1
            *command.tokens,
            LiteralCommand(
                "",
                sampling_method=SamplingMethod.COMBINATORIAL,
            ),  # sentinel 2
        ]

        def get_sequence(commands: list[Command]) -> typing.Iterable[list[str]]:
            if len(commands) == 0:
                yield []
            else:
                first_command, rest = commands[0], commands[1:]
                if context.get_effective_sampling_method(first_command).is_nonfinite():
                    for rest_vals in get_sequence(rest):
                        val = command_collection.get_value(first_command)
                        if val:
                            yield [val] + rest_vals
                        else:
                            yield rest_vals
                else:
                    gen = context.generator_from_command(first_command)
                    for first_val in gen:
                        for rest_vals in get_sequence(rest):
                            yield [first_val] + rest_vals

        word_arrays = get_sequence(augmented_tokens)
        for word_arr in word_arrays:
            prompt = command.separator.join(word_arr)
            yield prompt.strip(command.separator)

    def _get_variant(
        self,
        variant_command: VariantCommand,
        context: SamplingContext,
    ) -> StringGen:
        if len(variant_command.variants) == 0:
            return

        seen = set()
        is_wildcard_variant = len(variant_command.values) == 1 and isinstance(
            variant_command.values[0],
            WildcardCommand,
        )

        if is_wildcard_variant:
            wildcard_command = cast(WildcardCommand, variant_command.values[0])
            wildcard_variant = wildcard_to_variant(
                wildcard_command,
                context=context,
                min_bound=variant_command.min_bound,
                max_bound=variant_command.max_bound,
                separator=variant_command.separator,
            )
            yield from self._get_variant(wildcard_variant, context)
        else:
            variant_command = variant_command.adjust_range()
            for bound in range(
                variant_command.min_bound,
                variant_command.max_bound + 1,
            ):
                for combo in variant_command.get_value_combinations(bound):
                    for prompt_arr in _combo_to_prompt(context, combo):
                        deduped_arr = _dedupe(prompt_arr)
                        correct_size = len(deduped_arr) == bound
                        if correct_size and deduped_arr not in seen:
                            seen.add(deduped_arr)
                            yield variant_command.separator.join(deduped_arr)

    def _get_wildcard(
        self,
        command: WildcardCommand,
        context: SamplingContext,
    ) -> StringGen:
        values = context.wildcard_manager.get_all_values(command.wildcard)
        if len(values) == 0:
            logger.warning(f"No values found for wildcard {command.wildcard}")

        for val in values:
            # Parse and generate prompts from wildcard value
            yield from context.sample_prompts(val)

    def _get_literal(
        self,
        command: LiteralCommand,
        context: SamplingContext,
    ) -> StringGen:
        yield command.literal



from __future__ import annotations

import dataclasses
from typing import Iterable

from dynamicprompts.commands import Command, LiteralCommand, SamplingMethod


@dataclasses.dataclass(frozen=True)
class SequenceCommand(Command):
    tokens: list[Command]
    separator: str = ""
    sampling_method: SamplingMethod | None = None

    def __len__(self) -> int:
        return len(self.tokens)

    def __getitem__(self, index: int) -> Command:
        return self.tokens[index]

    def __iter__(self) -> Iterable[Command]:
        return iter(self.tokens)

    def __post_init__(self):
        assert all(
            isinstance(t, Command) for t in self.tokens
        ), "All tokens must be Command instances"

    @classmethod
    def from_literals(
        cls,
        values: list[str | Command],
        *,
        separator: str = "",
        sampling_method: SamplingMethod | None = None,
    ) -> SequenceCommand:
        return SequenceCommand(
            tokens=[
                v
                if isinstance(v, Command)
                else LiteralCommand(str(v), sampling_method=sampling_method)
                for v in values
            ],
            separator=separator,
            sampling_method=sampling_method,
        )



from enum import Enum


class SamplingMethod(Enum):
    RANDOM = "random"
    COMBINATORIAL = "combinatorial"
    CYCLICAL = "cycle"

    def is_nonfinite(self):
        return self in NON_FINITE_SAMPLING_METHODS


NON_FINITE_SAMPLING_METHODS = {SamplingMethod.RANDOM, SamplingMethod.CYCLICAL}



from __future__ import annotations

from dynamicprompts.generators.promptgenerator import PromptGenerator


class DummyGenerator(PromptGenerator):
    def generate(
        self,
        template: str,
        num_images: int = 1,
    ) -> list[str]:
        return num_images * [template]
